# -*- coding: utf-8 -*-
"""
Contains the definition of the DecayItem class.
"""
from __future__ import unicode_literals
from __future__ import print_function

import logging

from decay.constants import STRENGTH_SMALL_ENOUGH_TO_BE_0

logger = logging.getLogger('decay.item')


class DecayItem(object):
    """
    Contains attributes particular to a decaying item.

    Attributes:
        decay_strength (float):
            the strength of the item (subject to decay) at time decay_time
        decay_time (int):
            the last time a strength has been computed.
    """
    def __init__(self, decay_strength=1.0, decay_time=None, *args, **kwargs):
        """
        Constructor.

        Arguments:
            decay_strength (float):
                the strength of the item (subject to decay) at time decay_time
            decay_time (int):
                the last time a strength has been computed.
        """
        super().__init__(*args, **kwargs)

        self.decay_strength = decay_strength
        self.decay_time = decay_time

    def __str__(self):
        """ Represent this object as a human-readable string. """
        if self.decay_time is None:
            return 'DecayItem(#Invalid)'
        else:
            return 'DecayItem(%.1f, %d)' % (
                self.decay_strength, self.decay_time)

    def __repr__(self):
        """ Represent this object as a python constructor. """
        return 'DecayItem(%r, %r)' % (
                self.decay_strength, self.decay_time)

    def set_decay(self, new_tic, new_strength):
        """ Changes the value of the strength. """
        if new_strength < STRENGTH_SMALL_ENOUGH_TO_BE_0:
            new_strength = 0.0
        elif new_strength > 1.0:
            new_strength = 1.0
        self.decay_time = new_tic
        self.decay_strength = new_strength
        return self.decay_strength



# -*- coding: utf-8 -*-
"""
Contains the definition of the NoDecay class.
"""
from __future__ import unicode_literals
from __future__ import print_function

import logging

from decay.kind import DecayType

logger = logging.getLogger('decay.linear')


class LinearDecay(DecayType):
    """
    This is a decay placeholder that does not change the strength.
    """

    def __init__(self, *args, **kwargs):
        """
        Constructor.

        Arguments:

        """
        super().__init__(decay_name='.decay.linear.', *args, **kwargs)

    def __str__(self):
        """ Represent this object as a human-readable string. """
        return 'LinearDecay()'

    def __repr__(self):
        """ Represent this object as a python constructor. """
        return 'LinearDecay()'

    def update_decay(self, item, host, tic):
        """
        Change the value of the strength based on the time passed since th
        last update.

        Args:
            item:
                the item whose strength we're updating.
            host:
                the way we should update the item.
            tic:
                time at which we should compute the strength.

        Returns:
            Computed strength
        """
        logger.log(1, "linear decay for item %s hosted in %s at time %r",
                   item, host, tic)

        # This is the time it has passed since last update (x).
        # We normalize it inside the 0.0 - 1.0 interval
        delta_t = float(tic - item.decay_time) / \
            host.decay_tics_to_0

        # The y is the strength that should be between 0.0 and 1.0.
        # Based on it we compute the time (x) at that point.
        # In this case they are equal.
        # prev_time = 1.0 - relation.strength

        # So we get the new time in this system of reference.
        # new_time = prev_time + delta_t

        # new_strength = 1.0 - 1.0 + relation.strength - delta_t
        # new_strength = relation.strength - delta_t

        # And, at this point, we compute the new strength.
        # Because this is linear and both go from 0 to 1
        # This is just a matter of reversing the value.
        return item.set_decay(tic, item.decay_strength - delta_t)



# -*- coding: utf-8 -*-
"""
Contains the definition of the  class.
"""
from __future__ import unicode_literals
from __future__ import print_function

import logging
import numpy as np

from ...kind import DecayType

logger = logging.getLogger('')


class SampleBasedDecay(DecayType):
    """
    A base class that describes the curve in terms of samples.

    Usually there will be 20 samples for which both x and y are given.
    To compute a value that falls between these two a linear interpolation
    is used.

    Implementation is expected to have two attributes:
    decay_x and decay_y
    """
    def __init__(self, *args, **kwargs):
        """
        Constructor.

        Arguments:

        """
        super().__init__(*args, **kwargs)

    def __str__(self):
        """ Represent this object as a human-readable string. """
        return 'SampleBasedDecay()'

    def __repr__(self):
        """ Represent this object as a python constructor. """
        return 'SampleBasedDecay()'

    def sample_count(self):
        """ Tell the number of samples in this class. """
        assert len(self.decay_x) == len(self.decay_y)
        return len(self.decay_x)

    def x_decay_interval(self, x):
        """ Get the index of an x interval based on the value of x. """
        return np.searchsorted(self.decay_x, x, side='right') - 1

    def y_decay_interval(self, y):
        """ Get the index of an y interval based on the value of y. """
        result = self.sample_count() - np.searchsorted(
            np.flip(self.decay_y), y, side='left')
        return result - 1 if result > 0 else result

    def x_decay_for_y(self, y):
        """ Compute the value of x for the given y. """
        y_interval = self.y_decay_interval(y)
        if y_interval >= self.sample_count()-1:
            return self.decay_x[self.sample_count() - 1]
        else:
            smaller_y = self.decay_y[y_interval+1]
            higher_y = self.decay_y[y_interval]
            smaller_x = self.decay_x[y_interval]
            higher_x = self.decay_x[y_interval+1]
            # (higher_y - y) / (higher_y - smaller_y) =
            #     (x - smaller_x) / (higher_x - smaller_x)
            # (higher_x - smaller_x) * (higher_y - y) /
            #     (higher_y - smaller_y) = (x - smaller_x)
            # ((higher_x - smaller_x) * (higher_y - y) /
            #     (higher_y - smaller_y)) + smaller_x = x
            x = ((higher_x - smaller_x) * (higher_y - y) /
                 (higher_y - smaller_y)) + smaller_x
            return x

    def y_decay_for_x(self, x):
        """ Compute the value of y for the given x. """
        x_interval = self.x_decay_interval(x)
        if x_interval >= self.sample_count()-1:
            return self.decay_y[self.sample_count() - 1]
        else:
            smaller_y = self.decay_y[x_interval+1]
            higher_y = self.decay_y[x_interval]
            smaller_x = self.decay_x[x_interval]
            higher_x = self.decay_x[x_interval+1]
            # (higher_y - y) / (higher_y - smaller_y) =
            #     (x - smaller_x) / (higher_x - smaller_x)
            # (higher_y - y) = (x - smaller_x) * (higher_y - smaller_y) /
            #     (higher_x - smaller_x)
            #  -y = - higher_y + (x - smaller_x) * (higher_y - smaller_y) /
            #     (higher_x - smaller_x)
            y = higher_y - (x - smaller_x) * (
                    (higher_y - smaller_y) / (higher_x - smaller_x)
            )
            return y

    def update_decay(self, item, host, tic):
        """
        Change the value of the strength based on the time passed since th
        last update.

        Args:
            item:
                the item whose strength we're updating.
            host:
                the way we should update the item.
            tic:
                time at which we should compute the strength.

        Returns:
            Computed strength
        """
        logger.log(1, "sample based decay for item %s hosted in %s at time %r",
                   item, host, tic)

        # This is the time it has passed since last update (x).
        # We normalize it inside the 0.0 - 1.0 interval
        delta_t = float(tic - item.decay_time) / \
            host.decay_tics_to_0

        # The y is the strength that should be between 0.0 and 1.0.
        # Based on it we compute the time (x) at that point.
        prev_time = self.x_decay_for_y(item.decay_strength)

        # So we get the new time in this system of reference.
        new_time = prev_time + delta_t

        # And, at this point, we compute the new strength.
        return item.set_decay(tic, self.y_decay_for_x(new_time))



# -*- coding: utf-8 -*-
"""
The type of decays provided by the library are available in a concise form
in the registry. The users of the library may add other types by
using the dedicated methods.
"""
from .decays import (
    NoDecay, LinearDecay, CircleDecay, CubeDecay, HalfSuddenDecay,
    InverseCircleDecay, InverseParabolaDecay,
    ParabolaDecay, SigmoidDecay, SquareRootDecay, SuddenDecay,
    ThreeSinesDecay
)
from .constants import (
    NO_DECAY, LINEAR_DECAY, CIRCLE_DECAY, CUBE_DECAY, HALF_SUDDEN_DECAY,
    INVERSE_CIRCLE_DECAY, INVERSE_PARABOLA_DECAY, PARABOLA_DECAY,
    SIGMOID_DECAY, SQUARE_ROOT_DECAY, SUDDEN_DECAY, THREE_SINES_DECAY,
)


class Registry(object):
    """
    Stores all types of decays known to the library.
    """
    def __init__(self):
        global registry
        super().__init__()

        assert registry is None
        registry = self

        self.id_to_type = {}
        self.name_to_type = {}
        self.name_to_id = {}

        self._lib_init()

    @staticmethod
    def instance():
        return registry

    def add_type(self, kind, numeric_id):
        self.id_to_type[numeric_id] = kind
        self.name_to_type[kind.decay_name] = kind
        self.name_to_id[kind.decay_name] = numeric_id

    def remove_type(self, kind, numeric_id):
        del self.id_to_type[numeric_id]
        del self.name_to_type[kind.decay_name]

    def _lib_init(self):
        self.add_type(NoDecay(), NO_DECAY)
        self.add_type(LinearDecay(), LINEAR_DECAY)
        self.add_type(CircleDecay(), CIRCLE_DECAY)
        self.add_type(CubeDecay(), CUBE_DECAY)
        self.add_type(HalfSuddenDecay(), HALF_SUDDEN_DECAY)
        self.add_type(InverseCircleDecay(), INVERSE_CIRCLE_DECAY)
        self.add_type(InverseParabolaDecay(), INVERSE_PARABOLA_DECAY)
        self.add_type(ParabolaDecay(), PARABOLA_DECAY)
        self.add_type(SigmoidDecay(), SIGMOID_DECAY)
        self.add_type(SquareRootDecay(), SQUARE_ROOT_DECAY)
        self.add_type(SuddenDecay(), SUDDEN_DECAY)
        self.add_type(ThreeSinesDecay(), THREE_SINES_DECAY)


registry = Registry()



# -*- coding: utf-8 -*-
"""
Contains the definition of the DecayType class.
"""
from __future__ import unicode_literals
from __future__ import print_function

import logging

logger = logging.getLogger('decay.type')


class DecayType(object):
    """
    This class represents the base for all kinds of decays.

    This class require the implementation to
    have am `update_decay()` method and to name the decay.

    Attributes:
        decay_name (str):
            a unique name for this type of decay
    """

    def __init__(self, decay_name, *args, **kwargs):
        """
        Constructor.

        Arguments:
            decay_name (str):
                a unique name for this type of decay
        """
        super().__init__(*args, **kwargs)
        self.decay_name = decay_name

    def __str__(self):
        """ Represent this object as a human-readable string. """
        raise NotImplementedError

    def __repr__(self):
        """ Represent this object as a python constructor. """
        raise NotImplementedError

    def update_decay(self, item, host, tic):
        """
        Change the value of the strength based on the time passed since the
        last update.

        Args:
            item:
                the item whose strength we're updating.
            host:
                the way we should update the item.
            tic:
                time at which we should compute the strength.

        Returns:
            Computed strength
        """
        raise NotImplementedError

    @property
    def decay_type_id(self):
        """ Returns a numeric id of this type. """
        from .registry import Registry
        return Registry.instance().name_to_id[self.decay_name]



# -*- coding: utf-8 -*-
"""
Contains the definition of the NoDecay class.
"""
from __future__ import unicode_literals
from __future__ import print_function

import logging

from decay.kind import DecayType

logger = logging.getLogger('decay.none')


class NoDecay(DecayType):
    """
    This is a decay placeholder that does not change the strength.
    """

    def __init__(self, *args, **kwargs):
        """
        Constructor.

        Arguments:

        """
        super().__init__(decay_name='.decay.none.', *args, **kwargs)

    def __str__(self):
        """ Represent this object as a human-readable string. """
        return 'NoDecay()'

    def __repr__(self):
        """ Represent this object as a python constructor. """
        return 'NoDecay()'

    def update_decay(self, item, host, tic):
        """
        Change the value of the strength based on the time passed since th
        last update.

        Args:
            item:
                the item whose strength we're updating.
            host:
                the way we should update the item.
            tic:
                time at which we should compute the strength.

        Returns:
            Computed strength
        """
        logger.log(1, "no-op decay for item %s hosted in %s at time %r",
                   item, host, tic)
        return item.decay_strength



# -*- coding: utf-8 -*-
"""
"""
from .no_decay import NoDecay
from .linear import LinearDecay
from .sample.circle import CircleDecay
from .sample.cube import CubeDecay
from .sample.half_sudden import HalfSuddenDecay
from .sample.inverse_circle import InverseCircleDecay
from .sample.inverse_parabola import InverseParabolaDecay
from .sample.parabola import ParabolaDecay
from .sample.sigmoid import SigmoidDecay
from .sample.square_root import SquareRootDecay
from .sample.sudden import SuddenDecay
from .sample.three_sine import ThreeSinesDecay



# -*- coding: utf-8 -*-
"""
Contains the definition of the Decay class.
"""
from __future__ import unicode_literals
from __future__ import print_function

import logging

from .item import DecayItem
from .host import DecayHost

logger = logging.getLogger('Decay')


class Decay(DecayHost, DecayItem):
    """
    An item who'se strength decays over time.

    This is the "batteries included" version.
    """

    def __init__(self, *args, **kwargs):
        """
        Constructor.

        Arguments are passed to implementation classes.
        """
        super().__init__(*args, **kwargs)

    def __str__(self):
        """ Represent this object as a human-readable string. """
        return 'Decay()'

    def __repr__(self):
        """ Represent this object as a python constructor. """
        return 'Decay()'



# -*- coding: utf-8 -*-
"""
Contains the definition of the
"""
from __future__ import unicode_literals
from __future__ import print_function

import logging

logger = logging.getLogger('decay.host')


class DecayHost(object):
    """
    Can host one or several decay items, and contains the characteristics
    of the decay (type, duration).

    The usual use case is a list of items that inherits from this class while
    the items inherit from DecayItem. Thus the type of decay and the time
    it takes for an item's strength to decay are set in a central place
    (this class). Each item can have their own values in their journey
    from 1.0 to 0.0

    Attributes:
        decay_type (DecayType):
            an instance of the DecayType class that describes the way
            values change over time.
        decay_tics_to_0 (int):
            Number of tics (where tics can be any unit of time)
            required for a strength to go from 1.0 to 0.0
    """
    def __init__(self, decay_type=None, decay_tics_to_0=None, *args, **kwargs):
        """ Constructor. """
        super().__init__(*args, **kwargs)
        self.decay_type = decay_type
        self.decay_tics_to_0 = decay_tics_to_0

    @property
    def decay_type_id(self):
        """ Returns a numeric id of the type. """
        return self.decay_type.decay_type_id



# -*- coding: utf-8 -*-
"""
"""
import logging

__author__ = "Nicu Tofan"
__package_name__ = "decay"
__copyright__ = "Copyright 2019, Nicu Tofan"
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Nicu Tofan"
__email__ = "nicu.tofan@gmail.com"
__package_url__ = 'https://github.com/pyl1b/%s' % __package_name__

# Strength below this value gets converted to 0.
STRENGTH_SMALL_ENOUGH_TO_BE_0 = 0.00001


NO_DECAY = 1
LINEAR_DECAY = 2
CIRCLE_DECAY = 3
CUBE_DECAY = 4
HALF_SUDDEN_DECAY = 5
INVERSE_CIRCLE_DECAY = 6
INVERSE_PARABOLA_DECAY = 7
PARABOLA_DECAY = 8
SIGMOID_DECAY = 9
SQUARE_ROOT_DECAY = 10
SUDDEN_DECAY = 11
THREE_SINES_DECAY = 12
